#include <stdio.h>
#include <string.h>
#include <math.h>
#include <alloca.h>

#include "brightness.h"
#include "lava.h"
#include "mbi5026.h"
#include "debug.h"

#define TWO_PI (2.0 * 3.14159265358979323846)

#define MAX_DOTS 32

#define MAX_SIZE 11
#define WEIGHT_SIZE 15
int weight[MAX_SIZE + 1][WEIGHT_SIZE] = {
//
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x000, 0x000, 0x000, 0x000, 0x000, 0x00, 0x00, 0x00, 0x00, 0x00 }, //
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x000, 0x000, 0x100, 0x000, 0x000, 0x00, 0x00, 0x00, 0x00, 0x00 }, //
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x000, 0x040, 0x100, 0x040, 0x000, 0x00, 0x00, 0x00, 0x00, 0x00 }, //
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x020, 0x080, 0x100, 0x080, 0x020, 0x00, 0x00, 0x00, 0x00, 0x00 }, //
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x040, 0x0a0, 0x100, 0x0a0, 0x040, 0x00, 0x00, 0x00, 0x00, 0x00 }, //
		{ 0x00, 0x00, 0x00, 0x00, 0x20, 0x080, 0x0c0, 0x100, 0x0c0, 0x080, 0x20, 0x00, 0x00, 0x00, 0x00 }, //
		{ 0x00, 0x00, 0x00, 0x00, 0x40, 0x0a0, 0x0d0, 0x100, 0x0d0, 0x0a0, 0x40, 0x00, 0x00, 0x00, 0x00 }, //
		{ 0x00, 0x00, 0x00, 0x20, 0x80, 0x0c0, 0x0e0, 0x100, 0x0e0, 0x0c0, 0x80, 0x20, 0x00, 0x00, 0x00 }, //
		{ 0x00, 0x00, 0x00, 0x40, 0xa0, 0x0d0, 0x0f0, 0x100, 0x0f0, 0x0d0, 0xa0, 0x40, 0x00, 0x00, 0x00 }, //
		{ 0x00, 0x00, 0x20, 0x80, 0xc0, 0x0e0, 0x100, 0x100, 0x100, 0x0e0, 0xc0, 0x80, 0x20, 0x00, 0x00 }, //
		{ 0x00, 0x00, 0x40, 0xa0, 0xd0, 0x0f0, 0x100, 0x100, 0x100, 0x0f0, 0xd0, 0xa0, 0x40, 0x00, 0x00 }, //
		{ 0x00, 0x20, 0x80, 0xc0, 0xe0, 0x100, 0x100, 0x100, 0x100, 0x100, 0xe0, 0xc0, 0x80, 0x20, 0x00 } //
};
#define WEIGHT_MID 7

#define MULTIPLIER_SCALE 0x100

/**
 * @param v the center of the lava dot position (0.0 .. 1.0 * LEDS)
 * @param led the led to calculate the value for (0 .. LEDS)
 * @return a multiplier for the given position (0 .. MULTIPLIER_SCALE)
 */
int lava_multiplier(double v, int led, int size) {
	double led_ = 1.0 * led;
	if (led_ <= v - WEIGHT_MID || led_ >= v + WEIGHT_MID) {
		return 0;
	}

	double remainder = v - floor(v);

	int i1 = (int) floor(v) - led + WEIGHT_MID;
	int i2 = (int) floor(v + 1.0) - led + WEIGHT_MID;

	double multiplier = weight[size][i1] * (1.0 - remainder) + weight[size][i2] * remainder;
	return (int) floor(multiplier);
}

void lava_dot(struct data_lava_dot* dot, struct frame* frame) {
	memset(frame, 0, sizeof(struct frame));

	double position = sin(TWO_PI * dot->cursor / dot->period);
	position = (position + 1.0) / 2.0 * LEDS;

	for (int i = 0; i < LEDS; i++) {
		int m = lava_multiplier(position, i, dot->size);

		for (int c = 0; c < COLUMNS; c++) {
			if (dot->segments & (0x1 << c)) {
				frame->columns[c].leds[i].red = m * dot->red / MULTIPLIER_SCALE;
				frame->columns[c].leds[i].green = m * dot->green / MULTIPLIER_SCALE;
				frame->columns[c].leds[i].blue = m * dot->blue / MULTIPLIER_SCALE;
			}
		}
	}

	if (++dot->cursor >= dot->period) {
		dot->cursor = 0;
	}
}

void lava_message(struct data_lava* data, struct data_lava_dot* dots, int* num_dots) {
	switch (data->lava_command) {
	case L_ADD_DOT:
		//printf("Adding Lava dot %d\n", data->dot.id);
		memcpy(&dots[*num_dots], &data->dot, sizeof(struct data_lava_dot));
		(*num_dots)++;
		break;

	case L_UPDATE_DOT:
		//printf("Updating Lava dot %d\n", data->dot.id);
		for (int i = 0; i < *num_dots; i++) {
			struct data_lava_dot* dot = &dots[i];
			if (dot->id != data->dot.id) {
				continue;
			}

			dot->size = data->dot.size;
			dot->red = data->dot.red;
			dot->green = data->dot.green;
			dot->blue = data->dot.blue;
			if (dot->period != data->dot.period) {
				int new_cursor = dot->cursor * data->dot.period / dot->period;
				dot->period = data->dot.period;
				dot->cursor = new_cursor;
			}
			dot->segments = data->dot.segments;
			break;
		}
		break;

	case L_REMOVE_DOT:
		//printf("Removing Lava dot %d\n", data->dot.id);
		for (int i = 0; i < *num_dots; i++) {
			struct data_lava_dot* dot = &dots[i];
			if (dot->id != data->dot.id) {
				continue;
			}

			if (i < *num_dots - 1) {
				memcpy(&dots[i], &dots[i + 1], sizeof(struct data_lava_dot) * (*num_dots - i - 1));
			}

			(*num_dots)--;
		}
		break;

	default:
		break;
	}
}

void lava(struct data_lava* data) {
	if (data->lava_command != L_START) {
		return;
	}

	debug_log("Starting Lava mode\n");

	struct message message;

	int num_dots = 0;
	struct data_lava_dot dot_data[MAX_DOTS];
	struct frame dot_frames[MAX_DOTS];

	struct frame frame;

	for (;;) {
		for (int i = 0; i < num_dots; i++) {
			lava_dot(&dot_data[i], &dot_frames[i]);
		}

		if (num_dots > 0) {
			memcpy(&frame, &dot_frames[0], sizeof(struct frame));
			for (int i = 1; i < num_dots; i++) {
				add_frames(&dot_frames[i], &frame, &frame);
			}
		} else {
			memset(&frame, 0, sizeof(struct frame));
		}

		mbi_write_frame_pwm(&frame, 4);

		if (control_read(&message)) {
			switch (message.type) {
			case T_RESET:
				debug_log("Resetting from Lava mode\n");
				memset(&frame, 0, sizeof(struct frame));
				mbi_write_frame(&frame);
				return;
			case T_BRIGHTNESS:
				brightness(&message.data.brightness);
				break;
			case T_LAVA:
				lava_message(&message.data.lava, dot_data, &num_dots);
				break;
			default:
				break;
			}
		}
	}
}
