#include <stdlib.h>
#include <stdio.h>

#include "wiringPi.h"
#include "mbi5026.h"

int PIN_SDI[] = { 21, 20, 16, 12 };

#define PWM_ROWS 16
#define PWM_COLS 15

#define PWM_DELAY_MICROS 20

int pwm_matrix[PWM_ROWS][PWM_COLS] = {
/*
 { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 0
 { 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 1
 { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 }, // 2
 { 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0 }, // 3
 { 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1 }, // 4
 { 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1 }, // 5
 { 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1 }, // 6
 { 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0 }, // 7
 { 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1 }, // 8
 { 1, 1, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1, 0 }, // 9
 { 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0 }, // a
 { 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 0 }, // b
 { 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1 }, // c
 { 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1 }, // d
 { 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 }, // e
 { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 }  // f
 */
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 0
		{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 1
		{ 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 2
		{ 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 3
		{ 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 4
		{ 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 5
		{ 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 6
		{ 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0 }, // 7
		{ 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0 }, // 8
		{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0 }, // 9
		{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 }, // a
		{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0 }, // b
		{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0 }, // c
		{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0 }, // d
		{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 }, // e
		{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 }  // f
};

void mbi_init() {
	int rc;

	rc = wiringPiSetupGpio();
	if (rc != 0) {
		printf("Error setting up wiringPi");
		exit(EXIT_FAILURE);
	}

	for (int i = 0; i < COLUMNS; i++) {
		pinMode(PIN_SDI[i], OUTPUT);
		digitalWrite(PIN_SDI[i], 0);
	}

	pinMode(PIN_CLK, OUTPUT);
	digitalWrite(PIN_CLK, 0);

	pinMode(PIN_LE, OUTPUT);
	digitalWrite(PIN_LE, 0);

	pinMode(PIN_PWM, PWM_OUTPUT);
	pwmWrite(PIN_PWM, 0);
}

#define RAW_CELLS 4
#define RAW_LEDS RAW_CELLS * 16

struct raw_column {
	int value[RAW_LEDS];
};

struct raw_frame {
	struct raw_column columns[COLUMNS];
};


void mbi_transform_cell(struct column* column, struct raw_column* raw, int cell) {
	int i = LEDS / RAW_CELLS * cell;
	int j = RAW_LEDS / RAW_CELLS * (RAW_CELLS - 1 - cell);

	raw->value[j] = 0;
	raw->value[j + 7] = column->leds[i].red;
	raw->value[j + 9] = column->leds[i].green;
	raw->value[j + 8] = column->leds[i++].blue;
	raw->value[j + 4] = column->leds[i].red;
	raw->value[j + 5] = column->leds[i].green;
	raw->value[j + 6] = column->leds[i++].blue;
	raw->value[j + 1] = column->leds[i].red;
	raw->value[j + 2] = column->leds[i].green;
	raw->value[j + 3] = column->leds[i++].blue;
	raw->value[j + 12] = column->leds[i].red;
	raw->value[j + 11] = column->leds[i].green;
	raw->value[j + 10] = column->leds[i++].blue;
	raw->value[j + 15] = column->leds[i].red;
	raw->value[j + 14] = column->leds[i].green;
	raw->value[j + 13] = column->leds[i++].blue;
}

void mbi_transform_column(struct column* column, struct raw_column* raw) {
	for (int cell = 0; cell < RAW_CELLS; cell++) {
		mbi_transform_cell(column, raw, cell);
	}
}

void mbi_transform_frame(struct frame* frame, struct raw_frame* raw) {
	for (int column = 0; column < COLUMNS; column++) {
		mbi_transform_column(&frame->columns[column], &raw->columns[column]);
	}
}

void mbi_write_raw_frame(struct raw_frame* raw) {
	for (int i = 0; i < RAW_LEDS; i++) {
		for (int c = 0; c < COLUMNS; c++) {
			digitalWrite(PIN_SDI[c], raw->columns[c].value[i] > 0);
		}
		pulse(PIN_CLK);
	}
	pulse(PIN_LE);
}

void mbi_write_frame(struct frame* frame) {
	struct raw_frame raw;
	mbi_transform_frame(frame, &raw);

	mbi_write_raw_frame(&raw);
}

void mbi_write_raw_frame_pwm_iteration(struct raw_frame* raw, int pwm_col) {
	for (int i = 0; i < RAW_LEDS; i++) {
		for (int c = 0; c < COLUMNS; c++) {
			int pwm_row = (raw->columns[c].value[i] & 0xff) >> 4;
			digitalWrite(PIN_SDI[c], pwm_matrix[pwm_row][pwm_col]);
		}
		pulse(PIN_CLK);
	}

	pulse(PIN_LE);
}

void mbi_write_raw_frame_pwm(struct raw_frame* raw, int loops) {
	for (int l = 0; l < loops; l++) {
		for (int i = 0; i < PWM_COLS; i++) {
			mbi_write_raw_frame_pwm_iteration(raw, i);
			delayMicroseconds(PWM_DELAY_MICROS);
		}
	}
}

void mbi_write_frame_pwm(struct frame* frame, int loops) {
	struct raw_frame raw;
	mbi_transform_frame(frame, &raw);

	mbi_write_raw_frame_pwm(&raw, loops);
}

